/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Modified by the GLib Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/.
 */

#ifndef __G_LIB_H__
#define __G_LIB_H__

#define __GLIB_H_INSIDE__

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <float.h>
#include <assert.h>

#ifdef __cplusplus
extern "C"
#endif
{
	#define guint64 uint64_t
	#define guint32 uint32_t

	typedef char gchar;
	typedef void *gpointer;
	typedef const void *gconstpointer;
	typedef int gboolean;
	typedef unsigned int guint;
	typedef int gint;
	typedef char *GString;
	typedef guint32 gunichar;
	typedef unsigned long long gsize;
	typedef unsigned int	guint32;
	typedef unsigned char guchar;

	struct GMarkupParseContext;

	typedef enum
	{
		G_FILE_TEST_IS_REGULAR = 1 << 0,
		G_FILE_TEST_IS_SYMLINK = 1 << 1,
		G_FILE_TEST_IS_DIR = 1 << 2,
		G_FILE_TEST_IS_EXECUTABLE = 1 << 3,
		G_FILE_TEST_EXISTS = 1 << 4
	} GFileTest;

	struct GLibDelegate
	{
		void	*(*searchStart)(const char *dir);
		void	(*searchEnd)(void *searchStruct);
		const char	*(*searchNext)(void *searchStruct);
		
		int			(*fileTest)(const char *pathname, GFileTest test);
		
		int			(*matchWildCards)(const char *wildCardString, const char *compareString);
		
		int			(*parseXML)(GMarkupParseContext *context, const char *data, size_t dataSize);
		const char	*(*parseXMLGetElement)(GMarkupParseContext *context);

		int			(*parseLensName)(const char *model, float *minF, float *maxF, float *minA);
		
		const char	*(*readUTF8Char)(const char *ptr, unsigned int *uniChar);
		char			*(*casefoldUTF8)(const char *ptr, int length);
	};

	LF_EXPORT void g_set_delegate(GLibDelegate *del);

	struct GPtrArray
	{
		gpointer	*pdata;
		guint		len;
	};

	typedef struct _GError GError;

	class GQuark
	{
	};
	struct _GError
	{
		GQuark       domain;
		gint         code;
		gchar       *message;
	};

	#define G_FILE_ERROR g_file_error_quark()

	typedef enum
	{
		G_FILE_ERROR_EXIST,
		G_FILE_ERROR_ISDIR,
		G_FILE_ERROR_ACCES,
		G_FILE_ERROR_NAMETOOLONG,
		G_FILE_ERROR_NOENT,
		G_FILE_ERROR_NOTDIR,
		G_FILE_ERROR_NXIO,
		G_FILE_ERROR_NODEV,
		G_FILE_ERROR_ROFS,
		G_FILE_ERROR_TXTBSY,
		G_FILE_ERROR_FAULT,
		G_FILE_ERROR_LOOP,
		G_FILE_ERROR_NOSPC,
		G_FILE_ERROR_NOMEM,
		G_FILE_ERROR_MFILE,
		G_FILE_ERROR_NFILE,
		G_FILE_ERROR_BADF,
		G_FILE_ERROR_INVAL,
		G_FILE_ERROR_PIPE,
		G_FILE_ERROR_AGAIN,
		G_FILE_ERROR_INTR,
		G_FILE_ERROR_IO,
		G_FILE_ERROR_PERM,
		G_FILE_ERROR_NOSYS,
		G_FILE_ERROR_FAILED
	} GFileError;

	typedef enum
	{
		G_MARKUP_ERROR_BAD_UTF8,
		G_MARKUP_ERROR_EMPTY,
		G_MARKUP_ERROR_PARSE,
		/* The following are primarily intended for specific GMarkupParser
		* implementations to set.
		*/
		G_MARKUP_ERROR_UNKNOWN_ELEMENT,
		G_MARKUP_ERROR_UNKNOWN_ATTRIBUTE,
		G_MARKUP_ERROR_INVALID_CONTENT,
		G_MARKUP_ERROR_MISSING_ATTRIBUTE
	} GMarkupError;


	#ifndef FALSE
	#define FALSE (0)
	#endif

	#ifndef TRUE
	#define TRUE (!FALSE)
	#endif

	int g_parse_lens_name(const char *model, float &minf, float &maxf, float &mina);

	/* String/memory functions */
	LF_EXPORT gchar *g_strdup(const gchar *str);
	gchar *g_strndup(const gchar *str, int n);
	//gchar *g_strdup_printf(const gchar *format, ...) __attribute__((__format__(__printf__, 1, 2)));
	LF_EXPORT void *g_malloc(size_t size);
	LF_EXPORT void g_free(gpointer mem);
	LF_EXPORT void *g_realloc(void *, size_t);
	#define g_new(struct_type, n_structs) (struct_type *)g_malloc(sizeof(struct_type) * n_structs)

	char *g_markup_vprintf_escaped(const char *format, va_list args);
	void g_string_append(GString *output, const char *s);
	gunichar g_utf8_get_char(const char *c);
	const char *g_utf8_next_char(const char *c);
	int g_unichar_isspace(gunichar c);
	gunichar g_unichar_tolower(gunichar c);
	char *g_utf8_casefold(const char *c, int d);

	GPtrArray	*g_ptr_array_new();
	void g_ptr_array_add(GPtrArray *array, gpointer data);
	gpointer*  g_ptr_array_free(GPtrArray *array, gboolean free_seg);

	void g_ptr_array_set_size(GPtrArray *arr, guint length);
	#define g_ptr_array_index(arr,idx) (((arr)->pdata)[idx])
	//gpointer g_ptr_array_index(GPtrArray *arr, int index);
	void g_ptr_array_remove_range(GPtrArray *arr, guint index, guint index2);
	gpointer g_ptr_array_remove_index_fast(GPtrArray *array,	guint index_);

	gboolean g_str_has_prefix(const gchar *str, const gchar *prefix);
	GString *g_string_sized_new(gsize dfl_size);
	gchar *g_string_free(GString *string, gboolean free_segment);
	//struct_type *malloc_n((n_structs), sizeof(struct_type));
	//#define _G_NEW(struct_type, n_structs, func) ((struct_type *) ##func##_n ((n_structs), sizeof (struct_type)))
	//#define g_new(struct_type, n_structs) _G_NEW (struct_type, n_structs, malloc)
	//const lfCamera **g_new(lfCamera *, int);
	/* Hash functions */
	//#include "rbtree.h"

	typedef guint(*GHashFunc)(gconstpointer key);
	typedef gboolean(*GEqualFunc)(gconstpointer a, gconstpointer b);
	typedef gboolean(*GHRFunc)(gpointer key, gpointer value, gpointer user_data);
	typedef struct rb_root GHashTable;
	typedef void(*GDestroyNotify)(gpointer data);

	#define GUINT_TO_POINTER(a) ((void*)a)

	typedef gint(*GCompareFunc) (gconstpointer a, gconstpointer b);

	#define GDir	void

	GDir *g_dir_open(const char *, int, int *);
	const char *g_dir_read_name(GDir *dir);
	void g_dir_close(GDir *dir);

	char *g_build_filename(const char *, ...);
	LF_EXPORT void g_set_user_data_dir(const gchar *homeDir);
	const gchar *g_get_user_data_dir(void);
	gboolean g_file_get_contents(const gchar *filename, gchar **contents, gsize *length, GError **error);
	gboolean g_file_test(const gchar *filename, GFileTest test);
	int g_mkdir(const gchar *filename, int mode);
	int g_open(const gchar *filename, int flags, int mode);
	#ifndef PLATFORM_WINDOWS
		int _write(int _FileHandle, void const *_Buf, unsigned int _MaxCharCount);
		int _close(int _FileHandle);
		char *_strdup(const char *str);
	#endif

	#ifdef PLATFORM_WINDOWS
		#ifdef DEBUG
			#define g_warning printf
			#define g_assert assert
		#else
			#define g_warning(...)
			#define g_assert(a)
		#endif
	#else
		#ifdef DEBUG
			#define g_warning printf
			#define g_assert assert
		#else
			#define g_warning(...)
			#define g_assert(a)
		#endif
	#endif

	#define GStaticMutex int
	#define G_STATIC_MUTEX_INIT { NULL }
	//#define g_static_mutex_get_mutex g_static_mutex_get_mutex_impl
	int g_static_mutex_get_mutex(int *m);
	void g_mutex_lock(int v);
	#define g_static_mutex_lock(mutex) \
		 g_mutex_lock (g_static_mutex_get_mutex (mutex))
	void g_static_mutex_unlock(int *v);

	#define GPatternSpec char

	GPatternSpec *g_pattern_spec_new(const char *v);
	int g_pattern_match(GPatternSpec *pspec, guint string_length, const gchar *string, const gchar *string_reversed);
	void g_pattern_spec_free(GPatternSpec *pspec);

	void g_set_error(GError **err, GQuark domain, gint code,	const gchar *format,	...);

	#define G_MARKUP_ERROR g_markup_error_quark ()
	GQuark g_markup_error_quark(void);

	typedef enum
	{
		G_MARKUP_DO_NOT_USE_THIS_UNSUPPORTED_FLAG = 1 << 0,
		G_MARKUP_TREAT_CDATA_AS_TEXT = 1 << 1,
		G_MARKUP_PREFIX_ERROR_POSITION = 1 << 2,
		G_MARKUP_IGNORE_QUALIFIED = 1 << 3
	} GMarkupParseFlags;

	struct GMarkupParseContext;

	struct _GMarkupParser
	{
		/* Called for open tags <foo bar="baz"> */
		void(*start_element)  (GMarkupParseContext *context,
			const gchar         *element_name,
			const gchar        **attribute_names,
			const gchar        **attribute_values,
			gpointer             user_data,
			GError             **error);

		/* Called for close tags </foo> */
		void(*end_element)    (GMarkupParseContext *context,
			const gchar         *element_name,
			gpointer             user_data,
			GError             **error);

		/* Called for character data */
		/* text is not nul-terminated */
		void(*text)           (GMarkupParseContext *context,
			const gchar         *text,
			gsize                text_len,
			gpointer             user_data,
			GError             **error);

		/* Called for strings that should be re-saved verbatim in this same
		* position, but are not otherwise interpretable.  At the moment
		* this includes comments and processing instructions.
		*/
		/* text is not nul-terminated. */
		void(*passthrough)    (GMarkupParseContext *context,
			const gchar         *passthrough_text,
			gsize                text_len,
			gpointer             user_data,
			GError             **error);

		/* Called on error, including one set by other
		* methods in the vtable. The GError should not be freed.
		*/
		void(*error)          (GMarkupParseContext *context,
			GError              *error,
			gpointer             user_data);
	};

	typedef struct _GMarkupParser GMarkupParser;

	struct GMarkupParseContext
	{
		const GMarkupParser	*parser;
		gpointer					userData;
		void						*delegateData;
	};

	void g_markup_parse_context_get_position(GMarkupParseContext *context, gint *line_number, gint *char_number);
	const gchar *g_markup_parse_context_get_element(GMarkupParseContext *context);
	GMarkupParseContext *g_markup_parse_context_new(const GMarkupParser *parser, GMarkupParseFlags flags, gpointer user_data, GDestroyNotify user_data_dnotify);
	gboolean g_markup_parse_context_parse(GMarkupParseContext *context, const gchar *text, signed long long text_len,	GError **error);
	void  g_markup_parse_context_free(GMarkupParseContext *context);
#ifdef __cplusplus
}
#endif

#undef __GLIB_H_INSIDE__

#endif /* __G_LIB_H__ */
